import static org.junit.Assert.*;

import org.junit.Test;
import java.io.*;
import java.util.*;

public class TestVocabTrie
{
    // Small vocab file with only words (no counts or timestamps)
    private static final     String      WORDS_SMALL_FILENAME                = "words10.txt";
    private static final     int         WORDS_SMALL_SIZE                    = 10;
    private static final     String      WORDS_SMALL_FILENAME_COUNTS         = "words10_counts.txt";
    private static final     String      WORDS_SMALL_FILENAME_COUNTS_GZIP    = "words10_counts.txt.gz";
    private static final     String      WORDS_SMALL_FILENAME_COUNTS_ZIP     = "words10_counts.txt.zip";
    private static final     String      WORDS_SMALL_FILENAME_USED           = "words10_used.txt";    
    private static final     String      WORDS_SMALL_FILENAME_SORT           = "words10_sort.txt";    
    private static final     String []   WORDS10                             = {"the", "thee", "their", "this", "think", "thought", "throw", "three", "that", "that's"};

    // Big vocab file with only words (no counts or timestamps)
    private static final     String      WORDS_BIG_FILENAME                  = "wlist_match6.txt";
    private static final     int         WORDS_BIG_SIZE                      = 121932;
    private static final     String      WORDS_BIG_FILENAME_GZIP             = "wlist_match6.txt.gz";
    private static final     String      WORDS_BIG_FILENAME_ZIP              = "wlist_match6.txt.zip";

    private static final    String []   PREFIXES                            = {"the", "their", "this", "think", "these", "cat", "thin", "th", "t", "ze", "t'", "zzz"};


    // Randomly update the count and timestamp of words, write to disk, then see if we have same number of words.
    @Test
    public void testUpdateWordRandom() throws IOException
    {
        VocabTrie trie = new VocabTrie(WORDS_SMALL_FILENAME_USED, VocabTrie.FileType.PLAIN, VocabTrie.ResultType.ALL);

        // Randomly mark words as used
        for (int i = 0; i < 10000; i++)
        {
            int j = (int) (Math.random() * WORDS10.length);
            trie.updateWord(WORDS10[j]);
        }
        trie.write();

        // Load back in and make sure we have same number of words
        VocabTrie trie2 = new VocabTrie(WORDS_SMALL_FILENAME_USED, VocabTrie.FileType.PLAIN, VocabTrie.ResultType.ALL);
        assertEquals(trie.getSize(), trie2.getSize());
    }

    // Try and update some words that aren't in the vocab
    @Test
    public void testUpdateWordInvalid() throws IOException
    {
        VocabTrie trie = new VocabTrie(WORDS_SMALL_FILENAME_USED, VocabTrie.FileType.PLAIN, VocabTrie.ResultType.ALL);

        assertEquals(false, trie.updateWord("zebra"));
        assertEquals(false, trie.updateWord("theee"));
    }
    
    // Matching prefix words sorted by top count
    @Test
    public void testTop3Used() throws IOException
    {
        VocabTrie trie = new VocabTrie(WORDS_SMALL_FILENAME_SORT, VocabTrie.FileType.PLAIN, VocabTrie.ResultType.TOP3_USED);

        List<String> top3 = trie.getPrefixMatches("t");
        assertEquals(3, top3.size());       
        assertEquals("the", top3.get(0));
        assertEquals("thee", top3.get(1));
        assertEquals("their", top3.get(2));
    }

    // Matching prefix words sorted by top count, single matching word
    @Test
    public void testTop3UsedSingle() throws IOException
    {
        VocabTrie trie = new VocabTrie(WORDS_SMALL_FILENAME_SORT, VocabTrie.FileType.PLAIN, VocabTrie.ResultType.TOP3_USED);

        List<String> top3 = trie.getPrefixMatches("thee");
        assertEquals(1, top3.size());       
        assertEquals("thee", top3.get(0));
    }

    // Matching prefix words sorted by most recently used first
    @Test
    public void testTop3Recent() throws IOException
    {
        VocabTrie trie = new VocabTrie(WORDS_SMALL_FILENAME_SORT, VocabTrie.FileType.PLAIN, VocabTrie.ResultType.TOP3_RECENT);

        List<String> top3 = trie.getPrefixMatches("t");
        assertEquals(3, top3.size());       
        assertEquals("that's", top3.get(0));
        assertEquals("that", top3.get(1));
        assertEquals("thought", top3.get(2));
    }

    // Matching prefix words sorted by most recently used first, single matching word
    @Test
    public void testTop3RecentSingle() throws IOException
    {
        VocabTrie trie = new VocabTrie(WORDS_SMALL_FILENAME_SORT, VocabTrie.FileType.PLAIN, VocabTrie.ResultType.TOP3_RECENT);

        List<String> top3 = trie.getPrefixMatches("thought");
        assertEquals(1, top3.size());       
        assertEquals("thought", top3.get(0));
    }

    @Test
    public void testWriteSmallPlain() throws IOException
    {
        // Read in 10 words with count/timestamps, persist to disk
        VocabTrie trie = new VocabTrie(WORDS_SMALL_FILENAME_COUNTS, VocabTrie.FileType.PLAIN, VocabTrie.ResultType.ALL);
        trie.write();

        // Load back in and make sure count is the same
        VocabTrie trie2 = new VocabTrie(WORDS_SMALL_FILENAME_COUNTS, VocabTrie.FileType.PLAIN, VocabTrie.ResultType.ALL);
        assertEquals(trie.getSize(), trie2.getSize());
    }

    @Test
    public void testWriteSmallGzip() throws IOException
    {
        // Read in 10 words with count/timestamps, persist to disk
        VocabTrie trie = new VocabTrie(WORDS_SMALL_FILENAME_COUNTS_GZIP, VocabTrie.FileType.GZIP, VocabTrie.ResultType.ALL);
        trie.write();

        // Load back in and make sure count is the same
        VocabTrie trie2 = new VocabTrie(WORDS_SMALL_FILENAME_COUNTS_GZIP, VocabTrie.FileType.GZIP, VocabTrie.ResultType.ALL);
        assertEquals(trie.getSize(), trie2.getSize());
    }

    @Test
    public void testWriteSmallZip() throws IOException
    {
        // Read in 10 words with count/timestamps, persist to disk
        VocabTrie trie = new VocabTrie(WORDS_SMALL_FILENAME_COUNTS_ZIP, VocabTrie.FileType.ZIP, VocabTrie.ResultType.ALL);
        trie.write();

        // Load back in and make sure count is the same
        VocabTrie trie2 = new VocabTrie(WORDS_SMALL_FILENAME_COUNTS_ZIP, VocabTrie.FileType.ZIP, VocabTrie.ResultType.ALL);
        assertEquals(trie.getSize(), trie2.getSize());
    }   

    @Test
    public void testSizeBig() throws IOException
    {
        VocabTrie trie = getBigTrie();
        assertEquals(WORDS_BIG_SIZE, trie.getSize());
    }

    @Test
    public void testSizeBigGzip() throws IOException
    {
        VocabTrie trie = new VocabTrie(WORDS_BIG_FILENAME_GZIP, VocabTrie.FileType.GZIP, VocabTrie.ResultType.ALL);    
        assertEquals(WORDS_BIG_SIZE, trie.getSize());
    }

    @Test
    public void testSizeBigZip() throws IOException
    {
        VocabTrie trie = new VocabTrie(WORDS_BIG_FILENAME_ZIP, VocabTrie.FileType.ZIP, VocabTrie.ResultType.ALL);
        assertEquals(WORDS_BIG_SIZE, trie.getSize());
    }

    @Test
    public void testSizeSmall() throws IOException
    {
        VocabTrie trie = getSmallTrie();     
        assertEquals(WORDS_SMALL_SIZE, trie.getSize());
    }

    @Test
    public void testIsInVocab() throws IOException
    {
        VocabTrie small = getSmallTrie();

        assertEquals(true, small.isInVocab("the"));
        assertEquals(true, small.isInVocab("their"));
        assertEquals(true, small.isInVocab("this"));
        assertEquals(true, small.isInVocab("that's"));

        assertEquals(false, small.isInVocab("t"));
        assertEquals(false, small.isInVocab("th"));
        assertEquals(false, small.isInVocab("zebra"));
        assertEquals(false, small.isInVocab("'s"));
        assertEquals(false, small.isInVocab("thinking"));
    }

    @Test 
    public void testGetPrefixMatchesInvalidPrefix() throws IOException
    {
        VocabTrie small= getSmallTrie();

        List<String> matches = small.getPrefixMatches("zyq");
        assertEquals(0, matches.size());
    }

    @Test
    public void testGetPrefixMatchesString() throws IOException
    {
        VocabTrie big = getBigTrie();

        for (String prefix : PREFIXES)
        {
            List<String> matches = big.getPrefixMatches(prefix);

            // Make sure every returned word has the prefix at the start.
            // Also validate each match is in the vocab
            for (String match : matches)
            {
                assertEquals(0, match.indexOf(prefix));
                assertEquals(true, big.isInVocab(match));
            }
        }
    }

    private VocabTrie getSmallTrie() throws IOException
    {
        return new VocabTrie(WORDS_SMALL_FILENAME, VocabTrie.FileType.PLAIN, VocabTrie.ResultType.ALL);    
    }

    private VocabTrie getBigTrie() throws IOException
    {
        return new VocabTrie(WORDS_BIG_FILENAME, VocabTrie.FileType.PLAIN, VocabTrie.ResultType.ALL);    
    }

}
