/*
 * Stores a vocabulary of words in a prefix tree (trie).
 * This allows us to find all words that share a common prefix.
 * 
 * Author(s): Keith Vertanen
 */

import java.util.*;
import java.util.zip.*;
import java.io.*;

public class VocabTrie
{  
    // Different types of files we can load
    public enum FileType {PLAIN, GZIP, ZIP};

    // Different ways we can filter and/or sort the prefix matches
    public enum ResultType {ALL, TOP3_USED, TOP3_RECENT};
    
    // We don't know what subset of 16-bit unicode characters will be in the trie.
    // We maintain a map that takes any character and maps it to the an integer in [0..vocab size - 1].
    // This allows each node of the trie to be fast but reasonably memory compact.
    private Map<Character, Integer> mapCharToIndex = new HashMap<Character, Integer>();
    
    // We also want to quickly get from indexes back to characters
    private Map<Integer, Character> mapIndexToChar = new HashMap<Integer, Character>();
    
    private VocabTrieNode root = new VocabTrieNode();       // Root of the trie data structure
    private int realWords = 0;                              // How many actual real words are in the trie (not prefixes)
    
    private FileType fileType = FileType.PLAIN;             // What type of file we are support to read/write
    private String filename = "";                           // Filename we loaded from, also used to write any changes to
    private String innerFilename = "";                      // Filename inside a zip file if we loaded from one
    private ResultType resultType = ResultType.ALL;         // How we filter the prefix matches before returning to client
    
    // Represents a node in the trie, it has pointers (array) to all possible following letters.
    // Also a boolean flag to indicate if this path to the node represents a valid word and not 
    // just a prefix of a valid word (e.g. thei on the path to their).
    private class VocabTrieNode
    {
        private VocabTrieNode [] children;  // All the other prefixes/words leading from this node
        private VocabData data = null;      // If null, we are just a prefix and not a real word
    }

 
    // Create the trie based on a file that has a single word on each line
    public VocabTrie(String filename, FileType fileType, ResultType resultType) throws FileNotFoundException, IOException
    {
        List<String> lines = null;

        this.fileType = fileType;
        this.filename = filename;
        this.resultType = resultType;
        
        if (fileType == FileType.GZIP)
            lines = readLinesGzip(filename);        
        else if (fileType == FileType.ZIP)
            lines = readLinesZip(filename);        
        else if (fileType == FileType.PLAIN)
            lines = readLinesPlainText(filename);
        
        List<WordData> wordCountTimes = parseLines(lines);
        
        buildMap(wordCountTimes);
        
        for (WordData data : wordCountTimes)
            addWord(data);
    }
        
    // Given a sequence of lines read from disk, parse it out into 
    // the word string and then an optional count (integer) and timestamp (long).
    private List<WordData> parseLines(List<String> lines)
    {
        List<WordData> result = new ArrayList<WordData>();
        for (String line : lines)
        {
            String [] cols = line.split("\\s+");
 
            if (cols.length > 0)
            {
                // We should always have a word at the start of a line
                int count = 0;
                long time = 0;
                              
                // Optionally a line contains a count and Unix-style timestamp
                if (cols.length == 3)
                {
                    count = Integer.parseInt(cols[1]);
                    time = Long.parseLong(cols[2]);
                }
                WordData d = new WordData(cols[0], new VocabData(count, time));                
                result.add(d);
            }
        }        
        return result;
    }
      
    // Return all the lines from the given plain text filename.
    // We trim any whitespace and drop blank lines.
    private List<String> readLinesPlainText(String filename) throws FileNotFoundException, IOException
    {
        List<String> result = new ArrayList<String>();
        
        BufferedReader reader = new BufferedReader(new FileReader(filename));
        String line = reader.readLine();        
        while (line != null)
        {
            line = line.trim();
            if (line.length() > 0)
                result.add(line);
            line = reader.readLine();
        }
        reader.close();
        
        return result;
    }     
   
    // Return all the lines from the given gzipped text filename.
    // We trim any whitespace and drop blank lines.
    private List<String> readLinesGzip(String filename) throws FileNotFoundException, IOException
    {
        List<String> result = new ArrayList<String>();
        
        InputStream fileStream = new FileInputStream(filename);
        InputStream gzipStream = new GZIPInputStream(fileStream);
        Reader decoder = new InputStreamReader(gzipStream);
        BufferedReader reader = new BufferedReader(decoder);
        
        String line = reader.readLine();        
        while (line != null)
        {
            line = line.trim();
            if (line.length() > 0)
                result.add(line);
            line = reader.readLine();
        }
        reader.close();
        
        return result;
    }   
    
    // Return all the lines from the given zipped text filename.
    // Assumes there is only a single file in the *.zip file.
    // We trim any whitespace and drop blank lines.
    private List<String> readLinesZip(String filename) throws FileNotFoundException, IOException
    {
        List<String> result = new ArrayList<String>();
        
        InputStream     fileStream  = new FileInputStream(filename);
        ZipInputStream  zipStream   = new ZipInputStream(fileStream);
        Reader          decoder     = new InputStreamReader(zipStream);
        BufferedReader  reader      = new BufferedReader(decoder);
        
        // Advance stream to the start of the first file, remeber the filename since we'll need if we write to disk.
        ZipEntry entry = zipStream.getNextEntry(); 
        innerFilename = entry.getName();
        
        if (entry != null)   
        {
            
            String line = reader.readLine();        
            while (line != null)
            {
                line = line.trim();
                if (line.length() > 0)
                    result.add(line);
                line = reader.readLine();
            }
        }
        reader.close();
  
        return result;
    }   
        
    public void write()  throws FileNotFoundException, IOException
    {
        List<String> lines = getAllLines();

        if (fileType == FileType.GZIP)
            writeLinesGzip(filename, lines);        
        else if (fileType == FileType.ZIP)
            writeLinesZip(filename, lines);               
        if (fileType == FileType.PLAIN)
            writeLinesPlainText(filename, lines);                   
    }
    
    // Persist to disk using plain text file format
    private void writeLinesPlainText(String filename, List<String> lines) throws FileNotFoundException
    {
        PrintWriter writer = new PrintWriter(filename);
        for (String line : lines)
            writer.println(line);
        writer.close();
    }
    
    // Persist to disk using gzip file format
    private void writeLinesGzip(String filename, List<String> lines) throws FileNotFoundException, IOException
    {
        FileOutputStream output = new FileOutputStream(filename);
        GZIPOutputStream gzip = new GZIPOutputStream(output);
        Writer writer = new OutputStreamWriter(gzip);
        
        for (String line : lines)
            writer.write(line + "\n");
        writer.close();
    }
    
    // Persist to disk using zip file format
    private void writeLinesZip(String filename, List<String> lines) throws FileNotFoundException, IOException
    {
        StringBuilder sb = new StringBuilder();
        for (String line : lines)
            sb.append(line + "\n");

        File                file        = new File(filename);
        FileOutputStream    outStream   = new FileOutputStream(file);
        ZipOutputStream     zipStream   = new ZipOutputStream(outStream);
               
        ZipEntry entry = new ZipEntry(innerFilename);
        zipStream.putNextEntry(entry);
        
        byte [] data = sb.toString().getBytes();
        zipStream.write(data, 0, data.length);
        zipStream.closeEntry();
        zipStream.close();
    }
    
    // Returns a list with a line of text for each real word in the trie.
    // Each line contains the String of the word and the used count and last time stamp.
    private List<String> getAllLines()
    {
        List<String> result = new ArrayList<String>();        
        addAllLines(root, "", result);
        return result;
    }
    
    // Recursive helper method that goes through the whole trie
    private void addAllLines(VocabTrieNode node, String prefix, List<String> lines)
    {        
        // If this is a real word, add it to the running set of lines
        if (node.data != null)
        {
            String line = String.format("%s %d %d", prefix, node.data.getUsedCount(), node.data.getLastUsedTimestamp());
            lines.add(line);
        }
        
        if (node.children == null)
            return;
        
        // Recursively call ourselves on all children 
        for (int i = 0; i < node.children.length; i++)
        {
            if (node.children[i] != null)
                addAllLines(node.children[i], prefix + mapIndexToChar.get(i), lines);
        }
    }
    
    // Create the map that converts characters into an index into our fixed arrays at trie nodes.
    private void buildMap(List<WordData> wordCountTimes)
    {
        int i = 0;
        for (WordData data : wordCountTimes)
        {
            for (Character ch : data.getWord().toCharArray())
            {
                if (!mapCharToIndex.containsKey(ch))
                {
                    mapCharToIndex.put(ch, i);
                    mapIndexToChar.put(i, ch);
                    i++;
                }
            }
        }
    }
    
    // Helper method, add the specific word to the trie.
    // Maps each characters to the numeric LM ID.
    private void addWord(WordData wordData)
    {
        VocabTrieNode current = root;
        int vocabSize = mapCharToIndex.size();
        realWords++;
        
        for (int i = 0; i < wordData.getWord().length(); i++)
        {
            Character ch = wordData.getWord().charAt(i);
            int index = mapCharToIndex.get(ch);           
            if (current.children == null)
                current.children = new VocabTrieNode[vocabSize];
            if (current.children[index] == null)
                current.children[index] = new VocabTrieNode();
            current = current.children[index];
        }
        current.data = wordData.getData();
    }
    
    // See if the given word is in the vocabulary or not.
    public boolean isInVocab(String word)
    {
        List<Integer> ids = getWordIdsFromChracters(word);
        
        // If we can't map all the characters in the word to things in our vocab, then definitely not in vocab
        if (ids == null)
            return false;
        
        VocabTrieNode current = followPath(getWordIdsFromChracters(word));
        if (current == null)
            return false;
        return (current.data != null);
    }
    
    // Find all valid vocab words that have the given prefix.
    // This returns the words in String representation by converting back from numeric ID.
    // Returns an empty list if no matching words (or if prefix contains characters not in the vocab).
    public List<String> getPrefixMatches(String prefix)
    {
        List<String> result = new ArrayList<String>();
        
        List<Integer> ids = getWordIdsFromChracters(prefix);
        
        // If there are invalid characters in the prefix, then we return an empty list.
        if (ids == null)
            return result;
        
        VocabTrieNode current = followPath(ids);
        if (current == null)
            return result;
        
        // Recursively visit all nodes below this node and add to list.
        List<WordData> data = new ArrayList<WordData>();
        addChildren(current, prefix, data);
        
        // Now prune so we only return so many matches in a certain order
        if (resultType == ResultType.ALL)
            result = getWordsAll(data);
        else if (resultType == ResultType.TOP3_USED)
            result = getWordsTop3Used(data);
        else if (resultType == ResultType.TOP3_RECENT)
            result = getWordsTop3Recent(data);
        
        return result;
    }
    
    // Returns all the words in the matching vocab items. 
    // Does not sort in any particular way.
    private List<String> getWordsAll(List<WordData> data)
    {
        List<String> result = new ArrayList<String>();        
        for (WordData d : data)
            result.add(d.getWord());        
        return result;
    }
    
    // Out of all matching words, return the 3 with the highest counts.
    // Does not sort in any particular way.
    private List<String> getWordsTop3Used(List<WordData> data)
    {
        List<String> result = new ArrayList<String>();        
        Collections.sort(data, new WordData.CountComparator());
        for (int i = 0; i < Math.min(3, data.size()); i++)
            result.add(data.get(i).getWord());       
        return result;
    }
    
    // Out of all matching words, return the 3 with the highest counts.
    // Does not sort in any particular way.
    private List<String> getWordsTop3Recent(List<WordData> data)
    {
        List<String> result = new ArrayList<String>();        
        Collections.sort(data, new WordData.TimeComparator());
        for (int i = 0; i < Math.min(3, data.size()); i++)
            result.add(data.get(i).getWord());       
        return result;
    }
    
    // Convert the letters in a word into numeric indexes.
    // Returns null if any characters in str are not valid in this trie.
    private List<Integer> getWordIdsFromChracters(String str)
    {
        List<Integer> result = new ArrayList<Integer>();        
        for (int i = 0; i < str.length(); i++)
        {
            char ch = str.charAt(i);
            if (!mapCharToIndex.containsKey(ch))
                return null;            
            result.add(mapCharToIndex.get(ch)); 
        }      
        return result;
    }
    
    // Recursive helper method, adds to result variable all valid words below a certain node.
    private void addChildren(VocabTrieNode node, String prefix, List<WordData> result)
    {
        if (node.data != null)
        {
            WordData wordData = new WordData(prefix, node.data);
            result.add(wordData);
        }
        
        if (node.children == null)
            return;
        
        for (int i = 0; i < node.children.length; i++)
        {
            if (node.children[i] != null)
            {
                
                String extended = prefix + mapIndexToChar.get(i);
                addChildren(node.children[i], extended, result);
            }
        }
    }
    
    // Starting at the root, attempt to go down the give prefix path.
    // Returns null if we fail.
    private VocabTrieNode followPath(List<Integer> prefix)
    {
        VocabTrieNode current = root;
        
        // First find the starting node in the trie
        for (int i = 0; i < prefix.size(); i++)
        {
            int id = prefix.get(i);
            if (current.children == null)
                return null;
            if (current.children[id] == null)
                return null;
            current = current.children[id];
        }
        return current;
    }
 
    // Return the number of actual real words (not prefixes) stored in the trie
    public int getSize()
    {
        return realWords;
    }
    
    // Update our count of the number of times and timestamp of a word.
    // Returns true if we updated successfully, false otherwise (could be word is not in the vocab).
    public boolean updateWord(String word)
    {
       List<Integer> ids = getWordIdsFromChracters(word);
        
        // If we can't map all the characters in the word to things in our vocab, then definitely not in vocab
        if (ids == null)
            return false;
        
        VocabTrieNode current = followPath(getWordIdsFromChracters(word));
        if ((current == null) || (current.data == null))
            return false;
        
        current.data.used();
        return true;
    }
    
}
